#include "ringbuffer_internal.h"





/****************************************************************************/
void *ringbuffer_get_current_read_element( void *ringbuffer_state, void **ringbuffer_element )
{
  assert( ringbuffer_state != NULL );
  assert( ringbuffer_element != NULL );

  ringbuffer_internal_dequeue_element( ringbuffer_state, (struct ringbuffer_element **) ringbuffer_element );

  return( *ringbuffer_element );
}





/****************************************************************************/
void ringbuffer_get_current_write_element( void *ringbuffer_state, void **ringbuffer_element, int *overwrite_flag )
{
  struct ringbuffer_state
    *rs;

  void
    *freelist_element;

  assert( ringbuffer_state != NULL );
  assert( ringbuffer_element != NULL );
  assert( overwrite_flag != NULL );

  rs = (struct ringbuffer_state *) ringbuffer_state;

  /* TRD : we try to obtain an element from the freelist
           if we can, we populate it and add it to the queue

           if we cannot, then the ringbuffer is full
           so instead we grab the current read element and
           use that instead

           dequeue may fail since the queue may be emptied
           during our dequeue attempt

           so what we actually do here is a loop, attempting
           the freelist and if it fails then a dequeue, until
           we obtain an element

           once we have an element, we queue it

           you may be wondering why this operation is in a loop
           remember - these operations are lock-free; anything
           can happen in between

           so for example the pop could fail because the freelist
           is empty; but by the time we go to get an element from
           the queue, the whole queue has been emptied back into
           the freelist!
  */

  do
  {
    freelist_pop( rs->freelist_state, &freelist_element );

    if( freelist_element != NULL )
    {
      freelist_get_user_data_from_element( freelist_element, ringbuffer_element );
      (*(struct ringbuffer_element **) ringbuffer_element)->freelist_element = freelist_element;
    }

    if( freelist_element == NULL )
      ringbuffer_get_current_read_element( ringbuffer_state, ringbuffer_element );
  }
  while( *ringbuffer_element == NULL );

  if( freelist_element == NULL )
    *overwrite_flag = RAISED;

  if( freelist_element != NULL )
    *overwrite_flag = LOWERED;

  return;
}





/****************************************************************************/
void *ringbuffer_get_user_data_from_element( void *ringbuffer_element, void **user_data )
{
  struct ringbuffer_element
    *re;

  assert( ringbuffer_element != NULL );
  // TRD : user_data can be NULL

  re = (struct ringbuffer_element *) ringbuffer_element;

  if( user_data != NULL )
    *user_data = re->user_data;

  return( re->user_data );
}





/****************************************************************************/
void ringbuffer_set_user_data_in_element( void *ringbuffer_element, void *user_data )
{
  struct ringbuffer_element
    *re;

  assert( ringbuffer_element != NULL );
  assert( user_data != NULL );

  re = (struct ringbuffer_element *) ringbuffer_element;

  re->user_data = user_data;

  return;
}

