#include "slist_internal.h"





/****************************************************************************/
int slist_get_user_data_from_element( void *slist_element, void **user_data )
{
  struct slist_element
    *se;

  int
    rv = 1;

  assert( slist_element != NULL );
  assert( user_data != NULL );

  se = (struct slist_element *) slist_element;

  *user_data = (void *) se->user_data_and_flags[SLIST_USER_DATA];

  if( (atom_t) se->user_data_and_flags[SLIST_FLAGS] & SLIST_FLAG_DELETED )
    rv = 0;

  return( rv );
}





/****************************************************************************/
int slist_set_user_data_in_element( void *slist_element, void *user_data )
{
  struct slist_element
    *se;

  ALIGN(ALIGN_DOUBLE_POINTER) volatile void
    *user_data_and_flags[2],
    *new_user_data_and_flags[2];

  int
    rv = 1;

  assert( slist_element != NULL );
  // TRD : user_data can be NULL

  se = (struct slist_element *) slist_element;

  user_data_and_flags[SLIST_USER_DATA] = se->user_data_and_flags[SLIST_USER_DATA];
  user_data_and_flags[SLIST_FLAGS] = se->user_data_and_flags[SLIST_FLAGS];

  new_user_data_and_flags[SLIST_USER_DATA] = user_data;

  do
  {
    new_user_data_and_flags[SLIST_FLAGS] = user_data_and_flags[SLIST_FLAGS];
  }
  while( !((atom_t) user_data_and_flags[SLIST_FLAGS] & SLIST_FLAG_DELETED) and 0 == abstraction_dcas((volatile atom_t *) se->user_data_and_flags, (atom_t *) new_user_data_and_flags, (atom_t *) user_data_and_flags) );

  if( (atom_t) user_data_and_flags[SLIST_FLAGS] & SLIST_FLAG_DELETED )
    rv = 0;

  return( rv );
}





/****************************************************************************/
void *slist_get_head( void *slist_state, void **slist_element )
{
  struct slist_state
    *ss;

  assert( slist_state != NULL );
  assert( slist_element != NULL );

  ss = (struct slist_state *) slist_state;

  *slist_element = (void *) ss->head;

  slist_internal_move_to_first_undeleted_element( (struct slist_element **) slist_element );

  return( *slist_element );
}





/****************************************************************************/
void *slist_get_next( void *slist_element, void **next_slist_element )
{
  struct slist_element
    *se;

  assert( slist_element != NULL );
  assert( next_slist_element != NULL );

  se = (struct slist_element *) slist_element;

  *next_slist_element = (void *) se->next;

  slist_internal_move_to_first_undeleted_element( (struct slist_element **) next_slist_element );

  return( *next_slist_element );
}





/****************************************************************************/
void *slist_get_element_head_and_then_next( void *slist_state, void **slist_element )
{
  assert( slist_state != NULL );
  assert( slist_element != NULL );

  if( *slist_element == NULL )
    slist_get_head( slist_state, slist_element );
  else
    slist_get_next( *slist_element, slist_element );

  return( *slist_element );
}





/****************************************************************************/
void slist_internal_move_to_first_undeleted_element( struct slist_element **se )
{
  assert( se != NULL );

  while( *se != NULL and (atom_t) (*se)->user_data_and_flags[SLIST_FLAGS] & SLIST_FLAG_DELETED )
    (*se) = (struct slist_element *) (*se)->next;

  return;
}

