/***** the library wide include file *****/
#include "liblfds_internal.h"

/***** defines *****/
#define SLIST_USER_DATA  0
#define SLIST_FLAGS      1

#define SLIST_NO_FLAGS      0x0
#define SLIST_FLAG_DELETED  0x1

/***** incomplete types *****/
struct slist_element;

/***** structures *****/
#pragma pack( push, ALIGN_SINGLE_POINTER )

struct slist_state
{
  volatile struct slist_element
    *head;

  int
    (*user_data_delete_function)( void *user_data, void *user_state );

  void
    *user_state;
};

#pragma pack( pop )

#pragma pack( push, ALIGN_DOUBLE_POINTER )

/* TRD : this pragma pack doesn't seem to work under Windows
         if the structure members are the correct way round
         (next first), then user_data_and_flags ends up on
         a single pointer boundary and DCAS crashes

         accordingly, I've moved user_data_and_flags first
*/

struct slist_element
{
  volatile void
    *user_data_and_flags[2];

  volatile struct slist_element
    *next;
};

#pragma pack( pop )

/***** private prototypes *****/
void slist_internal_init_slist( struct slist_state *ss, int (*user_data_delete_function)(void *user_data, void *user_state), void *user_state );

void slist_internal_link_element_to_head( void *slist_state, volatile struct slist_element *se );
void slist_internal_link_element_after_element( volatile struct slist_element *slist_in_list_element, volatile struct slist_element *se );

void slist_internal_move_to_first_undeleted_element( struct slist_element **se );

