#include "slist_internal.h"





/****************************************************************************/
int slist_new( void **slist_state, int (*user_data_delete_function)(void *user_data, void *user_state), void *user_state )
{
  struct slist_state
    *ss;

  int
    rv = 0;

  assert( slist_state != NULL );

  ss = *slist_state = abstraction_aligned_malloc( sizeof(struct slist_state), ALIGN_SINGLE_POINTER );

  if( ss != NULL )
  {
    slist_internal_init_slist( ss, user_data_delete_function, user_state );
    rv = 1;
  }

  return( rv );
}





/****************************************************************************/
void slist_internal_init_slist( struct slist_state *ss, int (*user_data_delete_function)(void *user_data, void *user_state), void *user_state )
{
  assert( ss != NULL );
  // TRD : user_data_delete_function can be NULL
  // TRD : user_state can be NULL

  ss->head = NULL;
  ss->user_data_delete_function = user_data_delete_function;
  ss->user_state = user_state;

  return;
}





/****************************************************************************/
void *slist_new_head( void *slist_state, void *user_data )
{
  struct slist_state
    *ss;

  ALIGN(ALIGN_SINGLE_POINTER) volatile struct slist_element
    *se;

  assert( slist_state != NULL );
  // TRD : user_data can be NULL

  ss = (struct slist_state *) slist_state;
  se = abstraction_aligned_malloc( sizeof(struct slist_element), ALIGN_DOUBLE_POINTER );

  if( se != NULL )
  {
    se->user_data_and_flags[SLIST_USER_DATA] = user_data;
    se->user_data_and_flags[SLIST_FLAGS] = SLIST_NO_FLAGS;

    slist_internal_link_element_to_head( ss, se );
  }

  return( (void *) se );
}





/****************************************************************************/
void *slist_new_next( void *slist_element, void *user_data )
{
  ALIGN(ALIGN_SINGLE_POINTER) volatile struct slist_element
    *se;

  assert( slist_element != NULL );
  // TRD : user_data can be NULL

  se = abstraction_aligned_malloc( sizeof(struct slist_element), ALIGN_DOUBLE_POINTER );

  if( se != NULL )
  {
    se->user_data_and_flags[SLIST_USER_DATA] = user_data;
    se->user_data_and_flags[SLIST_FLAGS] = SLIST_NO_FLAGS;

    slist_internal_link_element_after_element( slist_element, se );
  }

  return( (void *) se );
}

