#include "internal.h"





/****************************************************************************/
void freelist_test( void )
{
  unsigned int
    loop,
    cpu_count;

  void
    *freelist_state;

  atom_t
    freelist_count;

  thread_state_t
    *thread_handles;

  /* TRD : tests

           1. one writer per CPU
              - freelist fills
           2. one reader per CPU
              - freelist empties
           3. one reader-writer per CPU
              - freelist balanced (we init with 10,000 elements)

           the first test is done by repeatedly calling freelist_new_elements()
           this function creates new elements and pushes them onto the freelist, ready for use

           the first test only runs for one second to avoid using up too much memory

           the second test them empties the freelist created in the first test

           be aware the second test memory leaks everything it pops :-)
  */

  cpu_count = abstraction_cpu_count();

  printf( "\n"
          "Freelist test\n"
          "=============\n" );

  // TRD : 1. one writer per CPU

  printf( "\n"
          "1. one writer per CPU\n"
          "=====================\n" );

  freelist_count = freelist_new( &freelist_state, 0, NULL, NULL );

  if( freelist_count != 0 )
  {
    printf( "freelist_count = %llu\n", (unsigned long long int) freelist_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], freelist_internal_thread_freelist_writer, freelist_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  // TRD : we use the freelist we made in the next test

  free( thread_handles );

  // TRD : 2. one reader per CPU

  printf( "\n"
          "2. one reader per CPU\n"
          "=====================\n" );

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], freelist_internal_thread_freelist_reader, freelist_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  freelist_delete( freelist_state, NULL, NULL );

  free( thread_handles );

  // TRD : 3. one reader-writer per CPU

  printf( "\n"
          "3. one reader-writer per CPU\n"
          "============================\n" );

  freelist_count = freelist_new( &freelist_state, 10000, NULL, NULL );

  if( freelist_count != 10000 )
  {
    printf( "freelist_count = %llu\n", (unsigned long long int) freelist_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], freelist_internal_thread_freelist_reader_writer, freelist_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  freelist_delete( freelist_state, NULL, NULL );

  free( thread_handles );

  return;
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION freelist_internal_thread_freelist_writer( void *freelist_state )
{
  time_t
    start_time;

  atom_t
    new_count;

  unsigned long long int
    count = 0;

  assert( freelist_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 1 )
  {
    new_count = freelist_new_elements( freelist_state, 1 );

    if( new_count == 1 )
      count++;
  }

  printf( "write count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION freelist_internal_thread_freelist_reader( void *freelist_state )
{
  void
    *freelist_element;

  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( freelist_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    freelist_pop( freelist_state, &freelist_element );

    if( freelist_element != NULL )
      count++;
  }

  printf( "read count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION freelist_internal_thread_freelist_reader_writer( void *freelist_state )
{
  void
    *freelist_element;

  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( freelist_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    freelist_pop( freelist_state, &freelist_element );

    if( freelist_element != NULL )
    {
      freelist_push( freelist_state, freelist_element );
      count++;
    }
  }

  printf( "write count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}

