#include "internal.h"





/****************************************************************************/
void ringbuffer_test( void )
{
  unsigned int
    loop,
    cpu_count;

  void
    *ringbuffer_state;

  atom_t
    ringbuffer_count;

  thread_state_t
    *thread_handles;

  /* TRD : there are five tests

           1. one reader thread per CPU
              - ringbuffer always empty
           2. one writer thread per CPU
              - ringbuffer always growing
           3. one reader and one writer thread per CPU
              - ringbuffer balanced
           4. two reader and one writer thread per CPU
              - ringbuffer tending to be empty
           5. one reader and two writers thread per CPU
              - ringbuffer tending to grow
  */

  // TRD : to get the number of CPUs
  cpu_count = abstraction_cpu_count();

  printf( "\n"
          "Ringbuffer test\n"
          "===============\n" );

  // TRD : 1. one reader thread per CPU

  printf( "\n"
          "1. one reader thread per CPU\n"
          "============================\n" );

  ringbuffer_count = ringbuffer_new( &ringbuffer_state, 10000, NULL, NULL );

  if( ringbuffer_state == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], ringbuffer_internal_thread_reader, ringbuffer_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( ringbuffer_state, NULL, NULL );

  free( thread_handles );

  // TRD : 2. one writer thread per CPU

  printf( "\n"
          "2. one writer thread per CPU\n"
          "============================\n" );

  ringbuffer_count = ringbuffer_new( &ringbuffer_state, 5, NULL, NULL );

  if( ringbuffer_state == NULL or ringbuffer_count != 5 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], ringbuffer_internal_thread_writer, ringbuffer_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( ringbuffer_state, NULL, NULL );

  free( thread_handles );

  // TRD : 3. one reader and one writer thread per CPU

  printf( "\n"
          "3. one reader and one writer thread per CPU\n"
          "===========================================\n" );

  ringbuffer_count = ringbuffer_new( &ringbuffer_state, 10000, NULL, NULL );

  if( ringbuffer_state == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 2 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], ringbuffer_internal_thread_reader, ringbuffer_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], ringbuffer_internal_thread_writer, ringbuffer_state );
  }

  for( loop = 0 ; loop < cpu_count * 2 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( ringbuffer_state, NULL, NULL );

  free( thread_handles );

  // TRD : 4. two reader and one writer thread per CPU

  printf( "\n"
          "4. two reader and one writer thread per CPU\n"
          "===========================================\n" );

  ringbuffer_count = ringbuffer_new( &ringbuffer_state, 10000, NULL, NULL );

  if( ringbuffer_state == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], ringbuffer_internal_thread_reader, ringbuffer_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], ringbuffer_internal_thread_reader, ringbuffer_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], ringbuffer_internal_thread_writer, ringbuffer_state );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( ringbuffer_state, NULL, NULL );

  free( thread_handles );

  // TRD : 5. one reader and two writers thread per CPU

  printf( "\n"
          "5. one reader and two writers thread per CPU\n"
          "============================================\n" );

  ringbuffer_count = ringbuffer_new( &ringbuffer_state, 10000, NULL, NULL );

  if( ringbuffer_state == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], ringbuffer_internal_thread_reader, ringbuffer_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], ringbuffer_internal_thread_writer, ringbuffer_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], ringbuffer_internal_thread_writer, ringbuffer_state );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( ringbuffer_state, NULL, NULL );

  free( thread_handles );

  return;
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION ringbuffer_internal_thread_reader( void *ringbuffer_state )
{
  void
    *ringbuffer_element;

  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( ringbuffer_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
    if( ringbuffer_get_current_read_element(ringbuffer_state, &ringbuffer_element) )
    {
      ringbuffer_put_read_element( ringbuffer_state, ringbuffer_element );
      count++;
    }

  printf( "read count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION ringbuffer_internal_thread_writer( void *ringbuffer_state )
{
  void
    *ringbuffer_element;

  time_t
    start_time;

  int
    overwrite_flag;

  unsigned long long int
    count = 0;

  assert( ringbuffer_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    ringbuffer_get_current_write_element( ringbuffer_state, &ringbuffer_element, &overwrite_flag );
    ringbuffer_put_write_element( ringbuffer_state, ringbuffer_element );

    count++;
  }

  printf( "write count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}

